

class SpikeStructLoadDataError(Exception):
    pass


class QualityNotLoadedError(Exception):
    def __str__(self):
        return (
            "to calculate quality measures, we recommend {}"
            " just add the quality.mat file to the working directory and retry.".format(
                SORTING_QUALITY_GITHUB_REPO
            )
        )


class SpikeTimeTypeError(Exception):
    pass


class StimulusTypeError(Exception):
    pass


class StimulusAlreadyPresentError(Exception):
    pass


class SubStimulusNotInRangeError(Exception):
    pass


class SubStimulusAlreadyPresentError(Exception):
    pass


class NoTracesInSpikeStructError(SpikeStructLoadDataError):
    def __str__(self):
        return "cannot find raw traces from which to find best waveform"


class IncorrectNchanTracesStructError(Exception):
    def __init__(self, n_datapoints, n_chan):
        super().__init__()
        self.n_dp = n_datapoints
        self.n_chan = n_chan

    def __str__(self):
        return (
            "n_chan incorrect, try again n_samples: {} divided "
            "by n_chan: {} not integer".format(self.n_dp, self.n_chan)
        )



import numpy as np
from cached_property import cached_property
from spike_handling import cluster_exceptions

from spike_handling.get_cluster_region_labels import get_cluster_region_label


class Cluster(object):
    def __init__(self, spike_struct, cluster_id):
        self.spike_struct = spike_struct
        self.cluster_id = cluster_id
        self.stimuli = []


    @property
    def group(self):
        """

        :return group: the group label assigned by manual clustering
        (usually done in phy: https://github.com/kwikteam/phy-contrib)
        """
        return self.spike_struct.groups[self.cluster_id]

    @property
    def best_channel_waveforms(self):
        """

        :return array best_channel_waveforms: every spike waveform for this cluster from the channel that is calculated
        as being closest to the source
        """
        return np.squeeze(self.waveforms[:, self.best_channel, :])

    @property
    def best_channel(self):
        """

        :return best_channel: the index of the channel that is calculated as being closest to the source
        """
        return self._get_channel_with_greatest_negative_deflection()

    def _get_channel_with_greatest_negative_deflection(self):
        _, min_channel = np.unravel_index(
            self.normalised_avg_waveforms.argmin(), self.avg_waveforms.shape
        )
        return min_channel

    @property
    def avg_waveforms(self):
        """

        :return avg_waveforms: the average waveform of the cluster's spikes, calculated for every channel and organised
         in a n_chan x t array where n_chan is the number of channels on the probe and t the number of samples in each
         waveform
        """
        return np.array(np.mean(self.waveforms, axis=2))

    @property
    def normalised_avg_waveforms(self):
        wfm = self.avg_waveforms
        med = np.median(wfm, axis=0)
        return wfm - med

    def _subtract_baseline_from_waveforms(self, waveforms):
        """

        :param waveforms:
        :return: median normalised waveforms
        """
        return waveforms - self.spike_struct.get_traces_median(
            n_samples=100000
        )

    @cached_property
    def waveforms(self, subtract_baseline=False, limit=100):
        """

        :param subtract_baseline: if the data are not normalised, some median correction needs to be applied to correct
        for the different baselines of the channels
        :return array waveforms: all spike waveforms across all channels and n_chan x t x n_waveforms
        """
        waveforms = self._get_spike_waveforms_on_all_channels(
            n_samples_before_peak=20, n_samples_after_peak=40, limit=limit
        )
        if subtract_baseline:
            waveforms = self._subtract_baseline_from_waveforms(waveforms)
        return waveforms

    def _get_spike_waveforms_on_all_channels(
        self, n_samples_before_peak, n_samples_after_peak, limit=None
    ):
        """

        :param int n_samples_before_peak:
        :param int n_samples_after_peak:
        :return np.array all_waveforms:
        """
        n_chan = self.spike_struct.n_chan
        spike_times = self.spike_times[:limit]
        n_waveforms = len(spike_times)
        n_waveform_samples = n_samples_before_peak + n_samples_after_peak

        all_waveforms = np.zeros((n_waveform_samples, n_chan, n_waveforms))

        for i, spike_time in enumerate(spike_times):
            if not (
                np.issubdtype(type(spike_time), np.int)
                or np.issubdtype(type(spike_time), np.uint)
            ):
                raise cluster_exceptions.SpikeTimeTypeError(
                    "got spike time:{}, type:{} expected "
                    "type: {}".format(spike_time, type(spike_time), int)
                )

            start_of_waveform = max(0, int(spike_time - n_samples_before_peak))
            end_of_waveform = int(
                spike_time + n_samples_after_peak
            )  # TODO min(len(trace), )

            waveform_on_all_channels = self.spike_struct.traces[
                start_of_waveform:end_of_waveform, :
            ]
            all_waveforms[:, :, i] = waveform_on_all_channels

        return all_waveforms

    def get_spike_times_in_interval(self, start, end):
        """

        :param start:
        :param end:
        :return spike_times: the time (n_samples) of all spikes that occur between the given start and end samples
        """
        spike_mask = np.logical_and(
            self.spike_times > start, self.spike_times < end
        )
        spike_times = self.spike_times[spike_mask]
        return spike_times

    @property
    def spike_times(self):
        """

        :return spike_times: the time of all spikes that belong to this cluster
        """
        return self.spike_struct.get_spike_times_in_cluster(self.cluster_id)

    def passes_criteria(self):
        (
            cluster_ids,
            _,
        ) = self.spike_struct.cluster_ids_that_pass_quality_measures()
        return self.cluster_id in cluster_ids

    def region_label(self, implant_distance=3000, tip_taper_distance=175):
        print(f"implant distance: {implant_distance}")
        print(f"tip taper distance: {tip_taper_distance}")
        cid, region_label, depth = get_cluster_region_label(
            self.spike_struct.mtd.histology_dir,
            self.spike_struct.mtd.kilosort_dir,
            self.cluster_id,
            implant_distance_um=implant_distance,
            tip_taper_distance=tip_taper_distance
        )
        return region_label, depth

    def __gt__(self, other):
        implant_distance=self.spike_struct.implant_distance
        return self.region_label(implant_distance)[1] < other.region_label(implant_distance)[1]

    def __lt__(self, other):
        implant_distance=self.spike_struct.implant_distance
        return self.region_label(implant_distance)[1] > other.region_label(implant_distance)[1]

    def curation_group(self):
        return self.spike_struct.groups[
            self.spike_struct.groups["cluster_id"] == self.cluster_id
        ]["group"].values[0]

    def metric_dict(self, implant_distance=3000):
        region, depth = self.region_label(implant_distance)
        metric_dict = {
            "mouse_id": self.spike_struct.mtd.mouse_id,
            "cluster_id": self.cluster_id,
            "passes_quality_criteria": self.passes_criteria(),
            "curation_group": self.curation_group(),
            "region": region,
            "depth": depth,
        }
        return metric_dict



import pandas as pd
import pathlib
import numpy as np

shank_mapping = {
    0: 0,
    32: 0,
    250: 1,
    282: 1,
    500: 2,
    532: 2,
    750: 3,
    782: 3,
}


def load_data(kilosort_output_directory):
    """load all kilosort numpy variables into a dictionary"""

    p = pathlib.Path(kilosort_output_directory)
    file_paths = list(p.rglob("*.npy"))
    return {p.stem: np.load(str(p)) for p in file_paths}


def get_best_channel(cluster_id, waveforms_folder):
    """Using output from quality measures, load waveforms and pick best channel for a given cluster"""

    waveforms_files = list(
        waveforms_folder.rglob("waveforms_{}.npy".format(cluster_id))
    )
    if len(waveforms_files) == 1:
        waveforms_file = waveforms_files[0]
    else:
        return None

    waveform = np.load(str(waveforms_file))
    avg_waveform = np.mean(waveform, axis=0)
    channel_mins = []
    for x in avg_waveform.T:
        channel_mins.append(np.min(x))

    return np.argmin(channel_mins)


def get_cluster_depth(cluster_id, data, kilosort_directory, shank_id=None):
    """use the estimation of the best channel to determine cluster location"""
    p = list(pathlib.Path(kilosort_directory).rglob("cluster_info.tsv"))
    phy_metadata = pd.read_table(str(p[0]))
    channel_id = phy_metadata[phy_metadata["cluster_id"] == cluster_id][
        "ch"
    ].values[0]
    if channel_id is None:
        return None, None
    channel_positions = data["channel_positions"]
    x, y = channel_positions[channel_id]
    if shank_id is None:
        shank_id = shank_mapping[x]
    depth = y
    return shank_id, depth


def get_region(track, depth, distance_key=None):
    """Use channel depth to select region from brainreg-segment data"""
    def find_nearest(array, value):
        array = np.asarray(array)
        idx = (np.abs(array - value)).argmin()
        return array[idx], idx

    if "Position" in track:
        distance_key = "Position"

    elif "Distance from first position [um]" in track:
        distance_key = "Distance from first position [um]"

    if distance_key is not None:
        distances = track[distance_key].values
        nearest_track_pos, nearest_track_pos_idx = find_nearest(distances, depth)
        region = track.iloc[nearest_track_pos_idx]["Region acronym"]

    else:
        print("couldn't find any position information for this track.. returning None")
        region = None
    return region


def get_cluster_region_label(
    brainreg_segment_dir,
    kilosort_folder,
    cluster_id,
    implant_distance_um=3000,
    surface=0,
    tip_taper_distance=175,
    shank_id=None,
):
    """use the channel position to select a brain region label for each cluster
    tracks must be drawn in the 'correct' order in brainreg-segment

    In phy 0 depth is the first channel. So we also need to add tip distance of 175um
    """
    data = load_data(kilosort_folder)
    p = pathlib.Path(brainreg_segment_dir)
    shank_id, depth = get_cluster_depth(
        cluster_id, data, kilosort_folder, shank_id=shank_id
    )

    if depth is not None:
        depth += tip_taper_distance

    if shank_id is None:
        return cluster_id, None, None

    print("cluster id: {}".format(cluster_id))
    print("shank id: {}".format(shank_id))
    print("cluster depth: {}".format(depth))

    this_track_path = list(p.rglob("track_{}.csv".format(shank_id)))
    assert len(this_track_path) == 1

    this_track = pd.read_csv(this_track_path[0])

    assert len(this_track) == implant_distance_um

    if surface == 0:
        depth = implant_distance_um - depth

    print(depth, implant_distance_um)

    if depth > 0:
        region = get_region(this_track, depth)
    else:
        region = None
    return cluster_id, region, depth



import matplotlib.pyplot as plt
import numpy as np
from scipy.io import loadmat
import cluster_exceptions as ce

from util import mat_utils

SAMPLE_RATE = 30000


class Stimulus(object):

    """
    locations of stimuli, and sub-stimuli to be interpreted by Cluster instances and used to extract spikes
    """

    def __init__(self, n_samples, wfm=None, condition=None, sp=None):
        self.start = 0
        self.end = n_samples
        self.relative_start = 0  # FIXME: probably can be removed
        self.relative_end = self.end - self.start
        self.sub_stimuli = []
        self.condition = condition
        self.offset = 0
        self.sp = sp
        self.duration_in_samples = self.end - self.start
        self.duration_in_seconds = self.duration_in_samples / SAMPLE_RATE

    @property
    def labels(self):
        return [sub.label for sub in self.sub_stimuli]

    def create_sub_stimulus(self, start, end, label):
        sub_stimulus = SubStimulus(start, end, label)
        self._append_sub_stimulus(sub_stimulus)

    def create_sub_stimuli(self, sub_region_starts, sub_region_ends, labels):
        for start, end, label in zip(
            sub_region_starts, sub_region_ends, labels
        ):
            sub_stimulus = SubStimulus(self, start, end, label)
            self._append_sub_stimulus(sub_stimulus)

    def _append_sub_stimuli(self, sub_stimuli):
        """

        :param list sub_stimuli:
        :return:
        """
        for sub_stimulus in sub_stimuli:
            self._append_sub_stimulus(sub_stimulus)

    def _append_sub_stimulus(self, sub_stimulus):
        if not self.relative_start <= sub_stimulus.start <= self.relative_end:
            raise ce.SubStimulusNotInRangeError(
                "sub_stimulus of start: {} and end: {} does not fit in "
                "range start: {} end: {}".format(
                    sub_stimulus.start,
                    sub_stimulus.end,
                    self.relative_start,
                    self.relative_end,
                )
            )

        if any([sub_stimulus == x for x in self.sub_stimuli]):
            raise ce.SubStimulusAlreadyPresentError(
                "sub-stimulus with start: {} and end: {} is already attached to "
                "this stimulus".format(sub_stimulus.start, sub_stimulus.end)
            )
            # TODO: more info in exception?

        self.sub_stimuli.append(sub_stimulus)

    def plot_waveform(self):
        plt.plot(self.wfm)

    def sort_sub_stimuli(self):
        self.sub_stimuli = np.sort(self.sub_stimuli)

    def __add__(self, other):
        # if not isinstance(other, int):
        #     raise TypeError('expected {} got {}'.format(int, type(other)))

        self.start += other
        self.end += other

        for sub_stimulus in self.sub_stimuli:
            sub_stimulus += other

    def reset(self):
        self.end -= self.start
        for sub_stimulus in self.sub_stimuli:
            sub_stimulus -= self.start
        self.start -= self.start


class SubStimulus(object):
    def __init__(self, stimulus, relative_start, relative_end, label=None):
        """

        :param stimulus_struct.stimulus:
        :param relative_start:
        :param relative_end:
        :param label:
        """
        self.start = relative_start
        self.end = relative_end
        self.duration_in_samples = relative_end - relative_start
        self.label = label
        self.stimulus = stimulus
        self.duration_in_seconds = self.duration_in_samples / SAMPLE_RATE

    def __lt__(self, other):
        return self.start < other.start

    def __gt__(self, other):
        return self.start > other.start

    def __eq__(self, other):
        if self.start == other.start and self.end == other.end:
            return True

    def __add__(self, other):
        self.start += other
        self.end += other

    def __sub__(self, other):
        self.start -= other
        self.end -= other

    def __same_group__(self, other):
        return np.isclose(
            self.duration_in_samples, other.duration_in_samples, rtol=1e-04
        )

    def get_data(self, cid, key):
        if key == "events":
            return self.events(cid)
        elif key == "n_events":
            return self.n_events(cid)
        elif key == "rate":
            return self.rate(cid)

    def events(self, cid):
        events = self.stimulus.sp.cluster_spike_times_in_interval(
            cid, self.start, self.end
        )
        return events

    def n_events(self, cid):
        return len(self.events(cid))

    def rate(self, cid):
        return self.n_events(cid) / self.duration_in_seconds

    @property
    def stripped_label(self):
        return "".join([i for i in self.label if not i.isdigit()])

    def plot(self, shift_factor=0):
        plt.plot(
            self.stimulus.waveform[self.start : self.end] + shift_factor,
            color="k",
            linewidth=0.5,
        )


class MateoStimulus(Stimulus):
    def __init__(self, n_samples, matlab_waveform_path=None, waveform=None):
        if waveform:
            self.waveform = waveform
        if matlab_waveform_path is not None:
            wfm = self.load_waveform_in_ms_from_matlab(matlab_waveform_path)
            self.waveform = self.upsample_waveform(wfm)
            n_samples = len(self.waveform)

        super().__init__(n_samples)

    def rotation_sub_stimuli(self, baseline_start=0, baseline_end=60000):
        peaks = mat_utils.findSinePeaks(self.waveform)
        cw1 = SubStimulus(peaks[0], peaks[1], "cw_1")
        acw1 = SubStimulus(peaks[1], peaks[2], "acw_1")
        bsl1 = SubStimulus(baseline_start, baseline_end, "bsl_1")
        cw2 = SubStimulus(peaks[2], peaks[3], "cw_2")
        acw2 = SubStimulus(peaks[3], peaks[4], "acw_2")
        bsl2 = SubStimulus(
            baseline_start + baseline_end, 2 * baseline_end, "bsl_2"
        )
        self._append_sub_stimuli([cw1, acw1, bsl1, cw2, acw2, bsl2])

    def stimulus_and_stationary_sub_stimuli(self):
        (
            rotation_start,
            rotation_end,
        ) = self.get_stimulus_boundaries_from_waveform()
        baseline_pre = SubStimulus(0, rotation_start, "baseline_pre")
        stimulus = SubStimulus(rotation_start, rotation_end, "stimulus")
        baseline_post = SubStimulus(
            rotation_end, len(self.waveform), "baseline_post"
        )
        self._append_sub_stimuli([baseline_pre, stimulus, baseline_post])

    def get_stimulus_boundaries_from_waveform(self):
        rotation_locations = np.where(self.waveform)[0]
        stimulus_start = rotation_locations[0]
        stimulus_end = rotation_locations[-1]
        return stimulus_start, stimulus_end

    @staticmethod
    def upsample_waveform(wfm, scale_factor=30):
        x = np.linspace(0, len(wfm) * scale_factor, len(wfm))
        y = wfm.flatten()
        x_vals = np.linspace(
            0, len(wfm) * scale_factor, len(wfm) * scale_factor
        )
        upsampled_command_waveform = np.interp(x_vals, x, y)
        return upsampled_command_waveform

    @staticmethod
    def load_waveform_in_ms_from_matlab(path):
        return loadmat(path)["waveform"]["data"][0][0]

    def _as_dictionary(self):
        import collections

        stim_dict = collections.OrderedDict()
        for sub in sorted(self.sub_stimuli):
            stim_dict[sub.label] = (sub.start, sub.end)
        return stim_dict


class SepiStimulus(Stimulus):
    def __init__(
        self,
        trigger,
        matlab_waveform_path=None,
        waveform=None,
        sp=None,
        condition=None,
    ):
        self.condition = condition
        self.matlab_waveform_path_template = matlab_waveform_path
        self.trigger = trigger

        if waveform:
            self.waveform = waveform

        if matlab_waveform_path is not None:
            wfm = self.load_waveform_in_ms_from_matlab(
                self.matlab_waveform_path
            )
            self.waveform = self.upsample_waveform(wfm)

        n_samples = len(self.waveform)
        super().__init__(n_samples, condition=condition, sp=sp)
        self.rotation_sub_stimuli()

    def rotation_sub_stimuli(self):
        peaks = mat_utils.findSinePeaks(self.waveform)
        (
            rotation_start,
            rotation_end,
        ) = self.get_stimulus_boundaries_from_waveform()
        cw = SubStimulus(self, peaks[0], peaks[1], "cw1")
        acw = SubStimulus(self, peaks[1], peaks[2], "acw1")
        baseline_pre_short = SubStimulus(
            self,
            rotation_start - cw.duration_in_samples,
            rotation_start,
            "baseline_pre_short",
        )
        baseline_post_short = SubStimulus(
            self,
            rotation_end,
            rotation_end + cw.duration_in_samples,
            "baseline_post_short",
        )

        baseline_pre = SubStimulus(self, 0, rotation_start, "baseline_pre")
        rotation = SubStimulus(self, rotation_start, rotation_end, "stimulus")
        baseline_post = SubStimulus(
            self, rotation_end, len(self.waveform), "baseline_post"
        )
        self._append_sub_stimuli(
            [
                baseline_pre_short,
                cw,
                acw,
                baseline_post_short,
                baseline_pre,
                rotation,
                baseline_post,
            ]
        )

    def stimulus_and_stationary_sub_stimuli(self):
        (
            rotation_start,
            rotation_end,
        ) = self.get_stimulus_boundaries_from_waveform()
        baseline_pre = SubStimulus(self, 0, rotation_start, "baseline_pre")
        stimulus = SubStimulus(self, rotation_start, rotation_end, "stimulus")
        baseline_post = SubStimulus(
            self, rotation_end, len(self.waveform), "baseline_post"
        )
        self._append_sub_stimuli([baseline_pre, stimulus, baseline_post])

    def get_stimulus_boundaries_from_waveform(self):
        rotation_locations = np.where(self.waveform)[0]
        stimulus_start = rotation_locations[0]
        stimulus_end = rotation_locations[-1]
        return stimulus_start, stimulus_end

    def substimuli_grouped_by_duration(self):
        stimuli_groups = {}
        for sub in self.sub_stimuli:
            if sub.duration_in_samples not in stimuli_groups.keys():
                stimuli_groups[sub.duration_in_samples] = []
            stimuli_groups[sub.duration_in_samples].extend(sub)
        return stimuli_groups

    @staticmethod
    def upsample_waveform(wfm, scale_factor=30):
        x = np.linspace(0, len(wfm) * scale_factor, len(wfm))
        y = wfm.flatten()
        x_vals = np.linspace(
            0, len(wfm) * scale_factor, len(wfm) * scale_factor
        )
        upsampled_command_waveform = np.interp(x_vals, x, y)
        return upsampled_command_waveform

    @staticmethod
    def load_waveform_in_ms_from_matlab(path):
        return loadmat(path)["waveform"]["data"][0][0]

    def _as_dictionary(self):
        import collections

        stim_dict = collections.OrderedDict()
        for sub in sorted(self.sub_stimuli):
            stim_dict[sub.label] = (sub.start, sub.end)
        return stim_dict

    def plot_sub_stimuli(self):
        self.plot_waveform(scale_factor=0.4)
        sub_stimulus_groups = self.substimuli_grouped_by_duration()
        for i, (duration, sub_stimuli) in enumerate(
            sub_stimulus_groups.items()
        ):
            for sub, color in zip(
                sub_stimuli, ["r", "b", "c", "g", "y"]
            ):  # FIXME: flexible color
                plt.hlines(
                    i, sub.start, sub.end, color=color, linewidth=20, alpha=0.4
                )

    def unique_labels(self):
        all_labels = np.array([sub.stripped_label for sub in self.sub_stimuli])
        _, idx = np.unique(all_labels, return_index=True)
        unique_labels_order_preserved = all_labels[np.sort(idx)]
        return unique_labels_order_preserved

    def plot_waveform(self, shift_factor=0, scale_factor=1.0, ax=None):
        x = np.arange(len(self.waveform)) + self.trigger
        if ax is None:
            ax = plt.subplot(111)
        ax.plot(
            x,
            self.waveform * scale_factor + shift_factor,
            color="k",
            linewidth=0.5,
        )

    def set_condition(self, condition):
        self.condition = condition

    @property
    def matlab_waveform_path(self):
        import re

        if self.condition is None:
            return None
        if self.condition == "None":
            return None
        rotation_angle = int(re.search(r"\d+", self.condition).group())
        waveform_filepath = self.matlab_waveform_path_template.format(
            rotation_angle
        )
        return waveform_filepath

    def get_data(self, cid, key):
        if key == "events":
            return self.events(cid)
        elif key == "n_events":
            return self.n_events(cid)
        elif key == "rate":
            return self.rate(cid)

    def events(self, cid):
        events = self.sp.cluster_spike_times_in_interval(
            cid, self.start, self.end
        )
        return events

    def n_events(self, cid):
        return len(self.events(cid))

    def rate(self, cid):
        return self.n_events(cid) / self.duration_in_seconds


class LoomStimulus(Stimulus):
    def __init__(self, loom_idx, sp=None, condition=None):
        self.condition = condition
        self.trigger = loom_idx[0]

        super().__init__(15 * 30000, condition=condition, sp=sp)
        self.loom_idx = loom_idx
        self.loom_sub_stimuli()

    def loom_sub_stimuli(self):
        start = 0
        stimulus_onset = 5 * 30000
        stimulus_offset = 10 * 30000
        end = 15 * 30000
        looms = [
            SubStimulus(
                self,
                loom - self.trigger + stimulus_onset,
                loom - self.trigger + 0.45 * 28000 + stimulus_onset,
                "loom{}".format(i),
            )
            for i, loom in enumerate(self.loom_idx)
        ]
        baseline_pre = SubStimulus(self, start, stimulus_onset, "baseline_pre")
        rotation = SubStimulus(
            self, stimulus_onset, stimulus_offset, "stimulus"
        )
        baseline_post = SubStimulus(
            self, stimulus_offset, end, "baseline_post"
        )
        self._append_sub_stimuli([baseline_pre, rotation, baseline_post])
        self._append_sub_stimuli(looms)

    def substimuli_grouped_by_duration(self):
        stimuli_groups = {}
        for sub in self.sub_stimuli:
            if sub.duration_in_samples not in stimuli_groups.keys():
                stimuli_groups[sub.duration_in_samples] = []
            stimuli_groups[sub.duration_in_samples].extend(sub)
        return stimuli_groups

    def _as_dictionary(self):
        import collections

        stim_dict = collections.OrderedDict()
        for sub in sorted(self.sub_stimuli):
            stim_dict[sub.label] = (sub.start, sub.end)
        return stim_dict

    def plot_sub_stimuli(self, relative=False):
        sub_stimulus_groups = self.substimuli_grouped_by_duration()
        for i, (duration, sub_stimuli) in enumerate(
            sub_stimulus_groups.items()
        ):
            for sub, color in zip(
                sub_stimuli, ["r", "b", "c", "g", "y"]
            ):  # FIXME: flexible color
                if "loom" in sub.label:
                    if relative:
                        plt.hlines(
                            i,
                            sub.start - 150000,
                            sub.end - 150000,
                            color=color,
                            linewidth=80,
                            alpha=0.4,
                        )
                    else:
                        plt.hlines(
                            i,
                            sub.start,
                            sub.end,
                            color=color,
                            linewidth=8000,
                            alpha=0.4,
                        )

    def unique_labels(self):
        all_labels = np.array([sub.stripped_label for sub in self.sub_stimuli])
        _, idx = np.unique(all_labels, return_index=True)
        unique_labels_order_preserved = all_labels[np.sort(idx)]
        return unique_labels_order_preserved

    def set_condition(self, condition):
        self.condition = condition

    def get_data(self, cid, key):
        if key == "events":
            return self.events(cid)
        elif key == "n_events":
            return self.n_events(cid)
        elif key == "rate":
            return self.rate(cid)

    def events(self, cid):
        events = self.sp.cluster_spike_times_in_interval(
            cid, self.start, self.end
        )
        return events

    def n_events(self, cid):
        return len(self.events(cid))

    def rate(self, cid):
        return self.n_events(cid) / self.duration_in_seconds



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from cached_property import cached_property
import stimulus_struct

from util import generic_functions

PLOT_SPACE = 5000
sns.set_color_codes("muted")

# TODO: make generic and extract rotation-based stuff to sub-class


class Stimuli(object):
    def __init__(
        self, trigger_indices, condition_list, sp, waveform_path_template
    ):
        """

        :param stimulus_struct.SepiStimulus stimulus:
        :param list trigger_indices: the sample numbers of all the triggers that you want to analyse
        """
        self.trigger_list = trigger_indices
        self.condition_list = condition_list
        self.waveform_path_template = waveform_path_template
        self.sp = sp

    def get_stimulus_template(self, condition):
        stimulus_template = stimulus_struct.SepiStimulus(
            0,
            matlab_waveform_path=self.waveform_path_template,
            sp=self.sp,
            condition=condition,
        )
        return stimulus_template

    @cached_property
    def stimuli(self):
        stimuli = []
        for trigger, condition in zip(self.trigger_list, self.condition_list):
            # print(condition)
            if condition == "None" or condition is None:
                continue
            stim = stimulus_struct.SepiStimulus(
                trigger,
                matlab_waveform_path=self.waveform_path_template,
                sp=self.sp,
                condition=condition,
            )
            stim + trigger
            stimuli.append(stim)
        return stimuli

    def get_event_data_all_sub_stimuli(
        self, cid, stimuli=None, relative=True, key="events"
    ):
        if stimuli is None:
            stimuli = self.stimuli

        # print('getting {} for cluster: {}'.format(key, cid))
        all_events = []
        labels = stimuli[0].unique_labels()

        for lbl in labels:
            sub_stimulus_events = []

            for stimulus in stimuli:
                for sub in stimulus.sub_stimuli:
                    events = sub.get_data(cid, key)

                    if relative:
                        events = events - sub.start

                    if lbl == sub.stripped_label:
                        sub_stimulus_events.append(events)

            all_events.append(sub_stimulus_events)

        return all_events, labels

    def get_rates_all_sub_stimuli(self, cid, stimuli=None):
        return self.get_event_data_all_sub_stimuli(
            cid=cid, stimuli=stimuli, relative=False, key="rate"
        )

    def get_n_events_all_sub_stimuli(self, cid, stimuli=None):
        return self.get_event_data_all_sub_stimuli(
            cid=cid, stimuli=stimuli, relative=False, key="n_events"
        )

    # def get_n_samples_from_label(self):  # FIXME
    #     for sub in self.stimulus_template.sub_stimuli:
    #         if sub.label == 'label':
    #             return sub.duration_in_samples

    def get_events_all_stimuli(
        self, cid, condition=None, relative=True, key="events"
    ):
        stimuli = (
            self.stimuli
            if condition is None
            else self.get_stimuli_in_condition(condition)
        )
        all_events = []
        if stimuli is None:
            stimuli = self.stimuli
        for stimulus in stimuli:
            events = stimulus.get_data(cid, key)
            if relative:
                events = events - stimulus.start
            all_events.append(events)
        return all_events

    def get_rates_all_stimuli(self, cid, condition=None):
        return self.get_events_all_stimuli(
            cid=cid, condition=condition, relative=False, key="rate"
        )

    def get_n_events_all_stimuli(self, cid, condition=None):
        return self.get_events_all_stimuli(
            cid=cid, condition=condition, relative=False, key="n_events"
        )

    def plot_all_stimuli(
        self,
        cid,
        condition=None,
        relative=True,
        key="events",
        fig=None,
        color=None,
    ):
        if fig is None:
            fig = plt.figure(facecolor="w", figsize=(15, 3))

        if condition is None:
            stimuli = self.stimuli
        else:
            stimuli = self.get_stimuli_in_condition(condition)

        stimulus_template = self.get_stimulus_template(stimuli[0].condition)

        all_events = self.get_events_all_stimuli(
            cid=cid, condition=condition, relative=relative, key=key
        )

        fig.add_subplot(211)
        plt.eventplot(all_events, colors=color, label=condition)
        stimulus_template.plot_waveform(shift_factor=len(stimuli) / 2)
        fig.axes[0].set_yticks([])

        fig.add_subplot(212)
        self.plot_histogram(
            all_events,
            duration_in_samples=stimulus_template.end,
            label="",
            color=color,
        )
        self.neaten_plots(fig.axes)
        self.neaten_plots(fig.axes[0:1], bottom=True, left=True)

    def plot_all_sub_stimuli(
        self,
        cid,
        condition=None,
        relative=True,
        fig=None,
        pyplot_override=False,
        n_bins=None,
    ):
        if fig is None:
            fig = plt.figure(facecolor="w", figsize=(30, 3))

        if condition is None:
            stimuli = self.stimuli

        else:
            stimuli = self.get_stimuli_in_condition(condition)

        event_lists, labels = self.get_event_data_all_sub_stimuli(
            cid=cid, stimuli=stimuli, relative=relative
        )
        histogram_axes = []
        raster_axes = []

        for i, (event_list, label) in enumerate(zip(event_lists, labels)):
            current_sub_stimulus = self.get_sub_stimulus(condition, label)

            ax = fig.add_subplot(2, len(event_lists), i + 1)
            self.sp.plot_raster(event_list, label=label)

            plt.xlim(
                [
                    0 - PLOT_SPACE,
                    current_sub_stimulus.duration_in_samples + PLOT_SPACE,
                ]
            )
            plt.ylim([-5, len(event_list) + 5])
            current_sub_stimulus.plot(shift_factor=len(event_list) / 2)
            raster_axes.append(ax)

            ax = fig.add_subplot(2, len(event_lists), i + 1 + len(event_lists))
            self.plot_histogram(
                event_list,
                duration_in_samples=current_sub_stimulus.duration_in_samples,
                label=label,
                pyplot_override=pyplot_override,
                n_bins=n_bins,
            )
            plt.xlim(
                [
                    0 - PLOT_SPACE,
                    current_sub_stimulus.duration_in_samples + PLOT_SPACE,
                ]
            )
            histogram_axes.append(ax)

        self.normalise_axes(histogram_axes, space=5)
        self.neaten_plots(fig.axes)
        self.neaten_plots(raster_axes, bottom=True)

    def get_sub_stimulus(self, condition, label):
        stimulus_template = self.get_stimulus_template(condition)
        for sub in stimulus_template.sub_stimuli:
            if label == sub.stripped_label:
                return sub

    @staticmethod
    def plot_raster(
        events, label=None
    ):  # TODO: extract for use in many modules
        plt.eventplot(events)
        if label is not None:
            plt.title(label)

    def plot_histogram(
        self,
        events,
        duration_in_samples,
        label=None,
        bin_duration_s=0.25,
        color=None,
        ax=None,
        scale_factor=1,
        stimulus_template=None,
        pyplot_override=False,
        n_bins=None,
    ):
        hist_events = np.array(generic_functions.flatten_list(events))
        if n_bins is None:
            n_bins = int(duration_in_samples / 30000 / bin_duration_s)
        bins = np.linspace(0, duration_in_samples, n_bins)
        hist, bin_edges = np.histogram(hist_events, bins=bins)
        bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2.0

        if ax is None:
            plt.step(bin_centers, hist / scale_factor, color=color)
            # plt.hist(hist_events, bins=bins, histtype='step', color=color)
        else:
            if pyplot_override:
                ax.hist(hist_events, bins=bins, histtype="step", color=color)
            else:
                ax.step(bin_centers, hist / scale_factor, color=color)
            if stimulus_template is not None:
                stimulus_template.plot_waveform(
                    shift_factor=np.median(hist) / scale_factor,
                    scale_factor=0.3,
                    ax=ax,
                )
            # ax.hist(hist_events, bins=bins, histtype='step', color=color)
        if label is not None:
            plt.title(label)

    @staticmethod
    def normalise_axes(axes, space=0):
        lower = min([ax.get_ylim()[0] for ax in axes])
        upper = max([ax.get_ylim()[1] for ax in axes])
        for ax in axes:
            ax.set_ylim([lower, upper + space])

    @staticmethod
    def neaten_plots(axes, top=True, right=True, left=False, bottom=False):
        for ax in axes:
            sns.despine(ax=ax, top=top, right=right, left=left, bottom=bottom)

    def get_stimuli_in_condition(self, condition):
        return [stim for stim in self.stimuli if stim.condition == condition]

    def get_events_all_stimuli_cluster_groups(self, cids, condition=None):
        from probez.util import generic_functions

        all_events = []
        for cid in cids:
            events = self.get_events_all_stimuli(cid, condition=condition)
            all_events.extend(generic_functions.flatten_list(events))
        return all_events

    def plot_data(
        self,
        cid,
        ax,
        plot_type="raster",
        condition=None,
        relative=True,
        key="events",
        color=None,
        scale_factor=1,
        n_bins=None,
        bin_duration_s=0.25,
    ):
        if condition is None:
            stimuli = self.stimuli
        else:
            stimuli = self.get_stimuli_in_condition(condition)

        stimulus_template = self.get_stimulus_template(stimuli[0].condition)
        all_events = self.get_events_all_stimuli(
            cid=cid, condition=condition, relative=relative, key=key
        )

        if plot_type == "raster":
            ax.eventplot(all_events, colors=color, label=condition)
            ax.set_yticks([])
            sns.despine(ax=ax, top=True, right=True, left=True, bottom=True)
            stimulus_template.plot_waveform(
                shift_factor=len(stimuli) / 2, ax=ax
            )

        elif plot_type == "histogram":
            self.plot_histogram(
                all_events,
                duration_in_samples=stimulus_template.end,
                label="",
                color=color,
                ax=ax,
                scale_factor=scale_factor,
                pyplot_override=True,
                n_bins=n_bins,
                bin_duration_s=bin_duration_s,
            )
            sns.despine(ax=ax, top=True, right=True)

    @staticmethod
    def calculate_stimuli_scale_ratio(stim_struct, condition, speed):
        """

        :return normalising_factor: normalising factor for different number of stimuli in light and dark for plotting
        """
        other = "dark {}" if condition == "light {}" else "light {}"

        stimuli_a = stim_struct.get_stimuli_in_condition(
            condition.format(speed)
        )
        stimuli_b = stim_struct.get_stimuli_in_condition(other.format(speed))
        arg = np.argmax([len(stimuli_a), len(stimuli_b)])
        if arg == 1:
            n_most_stimuli = max(len(stimuli_a), len(stimuli_b))
            n_fewest_stimuli = min(len(stimuli_a), len(stimuli_b))
            ratio = n_most_stimuli / n_fewest_stimuli
        else:
            ratio = 1

        return ratio

    def get_histogram_data(
        self,
        cid,
        bin_duration_s=0.25,
        scale_factor=1,
        n_bins=None,
        condition=None,
        relative=True,
        key="events",
    ):
        if condition is None:
            stimuli = self.stimuli
        else:
            stimuli = self.get_stimuli_in_condition(condition)

        stimulus_template = self.get_stimulus_template(stimuli[0].condition)

        all_events = self.get_events_all_stimuli(
            cid=cid, condition=condition, relative=relative, key=key
        )
        hist_events = np.array(generic_functions.flatten_list(all_events))
        if n_bins is None:
            n_bins = int(
                stimulus_template.duration_in_samples / 30000 / bin_duration_s
            )
        bins = np.linspace(0, stimulus_template.duration_in_samples, n_bins)
        hist, bin_edges = np.histogram(hist_events, bins=bins)
        bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2.0
        normalised_hist = hist / scale_factor
        return normalised_hist, bin_centers



import os
import pathlib
from bisect import bisect_left

import numpy as np
import pandas as pd
from cached_property import cached_property
from util import generic_functions
from spike_handling import waveforms
from spike_handling import cluster_exceptions
from spike_handling.cluster import Cluster


FS = 30000


class SpikeStruct(object):

    """
    class for loading and manipulating KiloSort output files and processed data

    example usage:
    >>> sp = SpikeStruct(directory)  # create the SpikeStruct and load all variables
    >>> good_clusters = sp.get_clusters_in_group('good')  # get all the clusters in a specific group
    >>> depth_ordered_good_clusters =  sp.sort_cluster_ids_by_depth(good_clusters, descend=True)  # order them


    """

    def __init__(self, session_mtd, sample_rate=30000, implant_distance=None):
        self.mtd = session_mtd
        self.sample_rate = sample_rate
        self.n_chan = 384
        self.traces_path = self.mtd.kilosort_dir / "temp_wh.dat"
        #self.unique_cluster_ids = np.unique(self.data["spike_times"])

        self.x_coords = self.data["channel_positions"][:, 0]
        self.y_coords = self.data["channel_positions"][:, 1]

        if implant_distance is not None:
            self.implant_distance = implant_distance

        self.groups = self.read_groups()
        if self.read_groups() is not None:
            if "KSLabel" in self.groups.keys():
                self.read_ks_uncurated_outputs()
            else:
                self.read_curated_ks_groups()

    def read_curated_ks_groups(self):
        grouplabel = "group"
        self.good_clusters = self.groups[
            self.groups[grouplabel] == "good"
            ]["cluster_id"].values
        self.MUA_clusters = self.groups[
            self.groups[grouplabel] == "mua"
            ]["cluster_id"].values
        self.noise_clusters = self.groups[
            self.groups[grouplabel] == "noise"
            ]["cluster_id"].values
        self.unsorted_clusters = self.groups[
            self.groups[grouplabel] == "unsorted"
            ]["cluster_id"].values
        self.other = set(
            np.unique(self.data["spike_times"])
        ).difference(set(self.groups["cluster_id"]))

    def read_ks_uncurated_outputs(self):
        grouplabel = "KSLabel"
        self.good_clusters = self.groups[
            self.groups[grouplabel] == "good"
            ]["cluster_id"].values
        self.MUA_clusters = self.groups[
            self.groups[grouplabel] == "mua"
            ]["cluster_id"].values
        self.noise_clusters = self.groups[
            self.groups[grouplabel] == "noise"
            ]["cluster_id"].values
        self.unsorted_clusters = self.groups[
            self.groups[grouplabel] == "unsorted"
            ]["cluster_id"].values

    def cluster_ids_that_pass_quality_measures(self, good_only=True):
        if good_only:
            cluster_ids = self.good_clusters
        else:
            cluster_ids = np.unique(
                np.concatenate([self.good_clusters, self.MUA_clusters])
            )
        (
            quality_cluster_ids,
            quality_df,
        ) = get_cluster_ids_that_pass_quality_criteria(self.mtd.quality_path)
        for cid in cluster_ids:
            filtered_cids = [
                cid for cid in cluster_ids if cid in quality_cluster_ids
            ]
        return filtered_cids, quality_df

    def bombcell_quality(self):
        bombcell_groups = pd.read_table(self.mtd.kilosort_dir / "cluster_bc_unitType.tsv")
        self.good_clusters = bombcell_groups[
            bombcell_groups['bc_unitType'] == 'GOOD'
            ]["cluster_id"].values
        self.MUA_clusters = bombcell_groups[
            bombcell_groups['bc_unitType'] == "MUA"
            ]["cluster_id"].values
        self.noise_clusters = bombcell_groups[
            bombcell_groups['bc_unitType'] == "NOISE"
            ]["cluster_id"].values
        self.non_somatic = bombcell_groups[
            bombcell_groups['bc_unitType'] == "NON-SOMA"
            ]["cluster_id"].values

    @cached_property
    def data(self):
        data = {
            p.stem: np.load(str(p))
            for p in self.mtd.kilosort_dir.rglob("*.npy")
        }
        data['spike_clusters'] = data['spike_clusters'].flatten()
        data['spike_times'] = data['spike_times'].flatten()
        return data

    @cached_property
    def traces(self):
        """
        Traces_path should be the path to the raw or processed binary data. This is used primarily for extracting
        waveforms so it helps if the data are high pass filtered or processed in some way, but it shouldn't be essential
        :param limit: restrict the size of the data used to improve speed
        :return shaped_data:
        """
        return load_traces(self.traces_path, self.n_chan)

    def read_groups(self):
        """
        manual sorting output (from e.g. phy) is stored as cluster_groups.csv
        :return dictionary manually_labelled_cluster_groups: a dictionary of group_ids for every cluster
        """
        path = os.path.join(self.mtd.kilosort_dir, "cluster_group.tsv")
        if not os.path.isfile(path):
            print("no cluster groups file")
            return None
        df = pd.read_csv(path, sep="\t")

        return df

    def get_clusters_in_group(self, group):
        """
        :param group: the group that the user wants clusters from
        :return clusters: a list of all cluster_ids classified to a user defined group:
        """
        if self.groups is None:
            return
        return [key for key in self.groups.keys() if self.groups[key] == group]


    def get_spike_times_in_interval(self, start_t, end_t, spike_times=None):
        """

        :param start_t: start point (n_samples)
        :param end_t: end point (n_samples)
        :param spike_times: the set of spikes from which to get subset from
        :return spike_times: all spike times within a user specified interval
        """
        if spike_times is None:
            spike_times = self.data["spike_times"]
        start_idx = bisect_left(spike_times, start_t)
        end_idx = bisect_left(spike_times, end_t)
        return spike_times[start_idx:end_idx]

    def get_spike_cluster_ids_in_interval(
        self, start_t, end_t, spike_times=None, spike_clusters=None
    ):
        """

        :param start_t: start point (n_samples)
        :param end_t: end point (n_samples)
        :param spike_clusters: the set of spikes from which to get subset from
        :param spike_times:
        :return spike_clusters: all cluster ids within a user specified interval
        """
        if spike_times is None:
            spike_times = self.data["spike_times"]
        if spike_clusters is None:
            spike_clusters = self.data["spike_times"]
        start_idx = bisect_left(spike_times, start_t)
        end_idx = bisect_left(spike_times, end_t)
        return spike_clusters[start_idx:end_idx]

    def get_spike_times_in_cluster(self, cluster_id):
        """

        :param int cluster_id:
        :return spike times: an array of all spike times within a user specified cluster
        """

        return self.data["spike_times"][
            self.spikes_in_cluster_mask(cluster_id).squeeze()
        ]

    def spikes_in_cluster_mask(self, cluster_id):
        """
        a boolean mask of all indices of spikes that belong to a group of user defined clusters

        :param int cluster_id:
        :return:
        """

        return self.data["spike_times"] == cluster_id

    def cluster_spike_times_in_interval(
        self, cluster_id, start, end, spike_times=None, spike_clusters=None
    ):
        """
        This function returns the spikes within an interval for a particular cluster id.
        NOTE the implementation is designed to allow arbitrary sets of spike times to be divided up, so doesnt
        always have to work from the full set of spike times and clusters.

        :param cluster_id:
        :param start:
        :param end:
        :param spike_times:
        :param spike_clusters:
        :return:
        """

        if spike_times is None:
            spike_times = self.data["spike_times"]
            spike_clusters = self.data["spike_clusters"]

        cluster_ids_in_interval, spike_times_in_interval = self.get_spikes_and_clusters_in_interval(start,
                                                                                                    end,
                                                                                                    spike_clusters,
                                                                                                    spike_times)

        this_cluster = cluster_ids_in_interval == cluster_id
        spikes_in_cluster_and_interval = spike_times_in_interval[
            this_cluster.squeeze()
        ]

        return spikes_in_cluster_and_interval

    def get_spikes_and_clusters_in_interval(self, start, end, spike_clusters, spike_times):
        spike_times_in_interval = self.get_spike_times_in_interval(
            start, end, spike_times
        )
        cluster_ids_in_interval = self.get_spike_cluster_ids_in_interval(
            start, end, spike_times, spike_clusters
        )
        return cluster_ids_in_interval, spike_times_in_interval

    def get_cluster_channel_from_avg_waveforms(self, cluster_id, n_spikes=100):
        """
        calculates the location of the cluster based on the greatest deflection from baseline of the average waveforms
        on all channels. channel*10 roughly corresponds to the depth of the channel on the probe

        :param cluster_id:
        :param n_spikes: number of spikes considered to form average waveform
        :return:
        """
        spike_times = self.get_spike_times_in_cluster(cluster_id)
        cluster_channel = waveforms.get_channel_of_max_amplitude_avg_waveform(
            spike_times[:n_spikes], self.traces, self.n_chan
        )

        return cluster_channel

    @cached_property
    def cluster_depths(self):
        df = pd.read_csv(
            pathlib.Path(self.mtd.kilosort_dir) / "cluster_info.tsv", sep="\t"
        )[["cluster_id", "depth", "group"]]

        return df

    def all_unsorted_cluster_ids(self):
        df = pd.read_csv(
            pathlib.Path(self.mtd.kilosort_dir) / "cluster_info.tsv", sep="\t"
        )[["cluster_id", "depth", "group"]]
        unsorted = df['cluster_id'].unique()
        noise = self.groups[self.groups['group'] == 'noise']['cluster_id'].unique()

        return set(unsorted) - set(noise)

    def get_clusters_in_depth_range(self, lower, upper, group_label="mua"):
        """
        :param cluster_ids:
        :param lower: the lower layer (channel) bound
        :param upper: the upper layer (channel) bound
        :return cluster_ids: a list of clusters within the two bounds
        """
        df = self.cluster_depths
        df_in_depth_boundaries = df[
            (df["depth"] > lower)
            & (df["depth"] < upper)
            & (df["group"] == group_label)
        ]
        return df_in_depth_boundaries["cluster_id"].unique()

    def sort_cluster_ids_by_depth(
        self, cluster_ids=None, cluster_depths=None, descend=True
    ):
        """
        :param cluster_ids:
        :param cluster_depths:
        :param descend: if descending order is required then set this to true
        :return:
        """
        if cluster_ids is None:
            cluster_ids = self.good_clusters
        if cluster_depths is None:
            cluster_depths = self.cluster_depths

        sorted_cluster_ids, cluster_depths = generic_functions.sort_by(
            cluster_ids, cluster_depths, descend=descend
        )
        return sorted_cluster_ids, cluster_depths

    def get_data_for_spikes_in_interval_and_unit(
        self, cluster_id, start, end, data_key="amplitudes"
    ):
        spike_time_mask = np.logical_and(
            self.data["spike_times"] > start, self.data["spike_times"] < end
        )
        cluster_id_mask = np.logical_and(
            self.data["spike_clusters"] == cluster_id, spike_time_mask
        )
        spike_data = self.data[data_key][cluster_id_mask]
        return spike_data

    def get_spike_amplitudes(self, cid, start, end):
        amplitudes = self.get_data_for_spikes_in_interval_and_unit(
            cid, start, end, "amplitudes"
        )
        return np.array(generic_functions.flatten_list(amplitudes))

    def clusters(self, cluster_ids=None):
        if cluster_ids is None:
            cluster_ids = self.cluster_ids_that_pass_quality_measures()[0]
        return [Cluster(self, cid) for cid in cluster_ids]

    def all_non_noise_clusters(self):
        cluster_ids= self.all_unsorted_cluster_ids()
        clusters =[Cluster(self, cid) for cid in cluster_ids]
        return clusters #sorted(clusters)

    def clusters_by_region(self, implant_distance, tip_taper_distance):
        cluster_region_dict = {}
        for c in self.all_non_noise_clusters():
            cluster_region_dict.setdefault(c.region_label(implant_distance=implant_distance,
                                                          tip_taper_distance=tip_taper_distance)[0], []).append(c.cluster_id)
        return cluster_region_dict


def load_traces(traces_path, n_chan):
    if not os.path.isfile(traces_path):
        raise cluster_exceptions.SpikeStructLoadDataError(
            "file: {} does not exist".format(self.traces_path)
        )
    data = np.memmap(traces_path, dtype=np.int16)
    if data.shape[0] % n_chan != 0:
        raise cluster_exceptions.IncorrectNchanTracesStructError(
            data.shape[0], self.n_chan
        )
    shape = (int(data.shape[0] / n_chan), n_chan)
    shaped_data = np.memmap(traces_path, shape=shape, dtype=np.int16)
    return shaped_data



def load_probe_data_memmap(path, n_chan):
    data = np.memmap(path, dtype="int16")
    if data.shape[0] % n_chan != 0:
        raise ValueError("n_chan is incorrect, try again")
    shape = (int(data.shape[0] / n_chan), n_chan)
    shaped_data = np.memmap(path, shape=shape, dtype="int16")
    return shaped_data, shape



import numpy as np


def get_spike_waveforms(
    spike_times,
    traces,
    n_chan,
    n_samples_before_peak=40,
    n_samples_after_peak=40,
):
    n_waveform_samples = n_samples_before_peak + n_samples_after_peak
    waveforms = np.zeros((n_waveform_samples, n_chan, len(spike_times)))

    for i, spike_time in enumerate(spike_times):
        spike_time = int(spike_time)
        start_of_waveform = int(spike_time - n_samples_before_peak)
        end_of_waveform = int(spike_time + n_samples_after_peak)
        waveform_on_all_channels = traces[start_of_waveform:end_of_waveform, :]
        waveforms[:, :, i] = waveform_on_all_channels
    return waveforms


def get_avg_waveforms(waveforms):
    """

    :param np.ndarray waveforms: all spike waveforms (list of np.arrays/memmaps across all channels)
    :return np.array avg_waveforms: the average waveforms on all channels
    """
    avg_waveforms = np.mean(waveforms, axis=2)
    return avg_waveforms


def get_channel_of_max_amplitude_avg_waveform(spike_times, traces, n_chan):
    all_waveforms = get_spike_waveforms(spike_times, traces, n_chan)
    avg_waveforms = get_avg_waveforms(all_waveforms)
    min_time_point, min_channel = np.unravel_index(
        avg_waveforms.argmin(), avg_waveforms.shape
    )
    return min_channel



import pandas as pd


def get_events(starts, ends, cids, sp):
    """

    :param starts:
    :param ends:
    :param cids:
    :param sp:
    :return:
    """
    histograms_all = []
    raster_all = [[] for x in range(len(starts))]
    for cid in cids:
        histogram = get_events_from_cluster(cid, starts, ends, sp, flat=True)
        raster = get_events_from_cluster(cid, starts, ends, sp, flat=False)
        spike_times = get_events_from_cluster(
            cid, starts, ends, sp, flat=True, relative=False
        )
        histograms_all.extend(histogram)
        for i, t in enumerate(raster):
            raster_all[i].extend(list(t))
    return histograms_all, raster_all, spike_times


def get_events_from_cluster(cid, starts, ends, sp, flat=False, relative=True, time=True):
    events = []
    for s, e in zip(starts, ends):
        subtract_start = s if relative else 0
        spikes_in_window = (
            sp.cluster_spike_times_in_interval(cid, s, e) - subtract_start
        )
        if time:
            if len(spikes_in_window) > 0:
                spikes_in_window = spikes_in_window / sp.sample_rate
        if flat:
            events.extend(spikes_in_window)
        else:
            events.append(spikes_in_window)

    return events


def get_trial_events(trial, sp, camera_frame_times_on_probe_samples, relative=True, time=True):
    s = trial.session.time_to_ephys(trial.start, camera_frame_times_on_probe_samples)
    e = s + (sp.sample_rate * 6)
    cluster_ids_in_interval, \
    spike_times_in_interval = sp.get_spikes_and_clusters_in_interval(s,
                                                                     e,
                                                                     sp.data["spike_clusters"],
                                                                     sp.data["spike_times"])
    subtract_start = s if relative else 0
    spikes_in_window = (
        spike_times_in_interval - subtract_start
    )
    if time:
        if len(spikes_in_window) > 0:
            spikes_in_window = spikes_in_window / sp.sample_rate
    data_dict = {
                 "cluster_id": cluster_ids_in_interval,
                 "spike_times": spikes_in_window,
    }
    df = pd.DataFrame.from_dict(data_dict)

    return df



import numpy as np


def flatten_list(lst):
    """
    this is the fastest
    """
    out = []
    for sublist in lst:
        out.extend(sublist)
    return out


def make_mask(size, idx_true=None):
    """
    Creates a boolean mask from a list of indices.

    :param size: The total size of the array.
    :param idx_true: A list of indices to be returned as True in the mask.
    :return: A boolean mask array of the specified size.
    """
    if idx_true is None:
        idx_true = list(range(size))

    mask = np.zeros(size, dtype=bool)
    mask[idx_true] = True
    return mask


def sort_by(list_to_sort, list_to_sort_by, descend=True):
    """
    sort one list by another list
    :param list list_to_sort:
    :param list list_to_sort_by:
    :param bool descend:
    :return list sorted_list:
    """

    sorted_lists = [
        (cid, did) for did, cid in sorted(zip(list_to_sort_by, list_to_sort))
    ]
    if descend:
        sorted_lists = sorted_lists[::-1]
    ordered = np.array(sorted_lists)[:, 0]
    ordered_by = np.array(sorted_lists)[:, 1]

    return list(ordered), list(ordered_by)


def rewrite_array_as_list_for_plotting(array):
    """function for restructuring data so it can be plotted more efficiently by matplotlib. Converts the input array
        into a list of lines separated by None values. NOTE: this means that all lines are together represented as a single
        line. Using this data structure to plot will, therefore, reduce the extent to which individual lines can be accessed
        by matplotlib.

    ==================   ==================
    line number              data
    ==================   ==================
    0                    1.8, 1.2, 1.6, 0.7
    1                    0.8, 1.2, 1.5, 0.9
    ==================   ==================

    will become

    ==================   ============================================
    line number              data
    ==================   ============================================
    0                    1.8, 1.2, 1.6, 0.7, None, 0.8, 1.2, 1.5, 0.9
    ==================   ============================================

    :param array array: the data you wish to plot
    :return x: the data as one None-separated line
    :return y: a list of indices to plot alongside
    """

    y = []
    x = []
    for item in array:
        y.extend(list(item))
        y.append(None)
        x.extend(list(range(len(item))))
        x.append(None)
    return x, y


